#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Cyborg v2 Acceleration accelerator action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import utils as oscutils
from oslo_serialization import jsonutils

from cyborgclient.common import utils
from cyborgclient import exceptions as exc
from cyborgclient.i18n import _


class ListDeviceProfile(command.Lister):
    """List all device profiles"""

    def get_parser(self, prog_name):
        parser = super(ListDeviceProfile, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("List additional fields in output")
        )
        return parser

    def take_action(self, parsed_args):
        acc_client = self.app.client_manager.accelerator

        if parsed_args.detail:
            column_headers = (
                "created_at",
                "updated_at",
                "uuid",
                "name",
                "groups"
            )
            columns = (
                "created_at",
                "updated_at",
                "uuid",
                "name",
                "groups"
            )
        else:
            column_headers = (
                "uuid",
                "name",
                "groups",
            )
            columns = (
                "uuid",
                "name",
                "groups",
            )

        data = acc_client.device_profiles()
        if not data:
            return (), ()
        formatters = {}
        return (column_headers,
                (oscutils.get_item_properties(
                    s, columns, formatters=formatters) for s in data))


class CreateDeviceProfile(command.ShowOne):
    """Register a new device_profile with the accelerator service"""
    log = logging.getLogger(__name__ + ".CreateDeviceProfile")

    def get_parser(self, prog_name):
        parser = super(CreateDeviceProfile, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_("Unique name for the device_profile."))
        parser.add_argument(
            'groups',
            metavar='<groups>',
            help=_("groups for the device_profile."))
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        acc_client = self.app.client_manager.accelerator

        attrs = {
            'name': parsed_args.name,
            'groups': list(jsonutils.loads(parsed_args.groups)),
        }
        device_profile = acc_client.create_device_profile(**attrs)
        return _show_device_profile(acc_client, device_profile.uuid)


class DeleteDeviceProfile(command.Command):
    """Delete deviceProfile(s)."""

    log = logging.getLogger(__name__ + ".DeleteDeviceProfile")

    def get_parser(self, prog_name):
        parser = super(DeleteDeviceProfile, self).get_parser(prog_name)
        parser.add_argument(
            "device_profiles",
            metavar="<uuid>",
            nargs="+",
            help=_("UUID(s) of the device_profile(s) to delete.")
        )

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        failures = []
        for uuid in parsed_args.device_profiles:
            try:
                acc_client.delete_device_profile(uuid, False)
                print(_('Deleted device_profile %s') % uuid)
            except exc.ClientException as e:
                failures.append(_("Failed to delete device_profile \
                                %(device_profile)s: %(error)s")
                                % {'uuid': uuid, 'error': e})
        if failures:
            raise exc.ClientException("\n".join(failures))


class ShowDeviceProfile(command.ShowOne):
    """Show device_profile details."""
    log = logging.getLogger(__name__ + ".ShowDeviceProfile")

    def get_parser(self, prog_name):
        parser = super(ShowDeviceProfile, self).get_parser(prog_name)
        parser.add_argument(
            "device_profile",
            metavar="<uuid>",
            help=_("UUID of the device_profile.")
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        acc_client = self.app.client_manager.accelerator
        return _show_device_profile(acc_client,
                                    parsed_args.device_profile)


def _show_device_profile(acc_client, uuid):
    """Show detailed info about device_profile."""
    columns = (
        "created_at",
        "updated_at",
        "uuid",
        "name",
        "groups",
    )
    try:
        device_profile = acc_client.get_device_profile(uuid)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('device_profile not found: %s') % uuid)
    formatters = {
        'data': utils.json_formatter,
    }
    data = device_profile.to_dict()
    return columns, oscutils.get_dict_properties(data, columns,
                                                 formatters=formatters)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testtools
from testtools import matchers

from cyborgclient.tests.unit import utils
from cyborgclient.v1 import accelerators

ACCELERATOR1 = {"user_id": "3a4b753552964978af7f76ce9fecf7d0",
                "description": "test", "links":
                [{"href": "http://127.0.0.1:6666/v1/accelerators/fake_uuid",
                          "rel": "self"},
                 {"href": "http://127.0.0.1:6666/accelerators/fake_uuid",
                  "rel": "bookmark"}],
                "acc_capability": "test",
                "created_at": "2018-07-20T07:45:04+00:00",
                "vendor_id": "test", "updated_at": None,
                "acc_type": None, "name": "test-cyborg-create",
                "product_id": "test", "device_type": "test", "remotable": 1,
                "project_id": None,
                "uuid": "4cc55aab-dac6-486f-ad14-284c8e554589"}

ACCELERATOR2 = {"user_id": "3a4b753552964978af7f76ce9fecf7d0",
                "description": "test", "links":
                [{"href": "http://127.0.0.1:6666/v1/accelerators/fake_uuid",
                  "rel": "self"},
                 {"href": "http://127.0.0.1:6666/accelerators/fake_uuid",
                  "rel": "bookmark"}],
                "acc_capability": "test",
                "created_at": "2018-07-20T08:11:14+00:00",
                "vendor_id": "test", "updated_at": None,
                "acc_type": None, "name": "test-cyborg-create",
                "product_id": "test", "device_type": "test", "remotable": 1,
                "project_id": None,
                "uuid": "a444397a-deba-4c94-984f-ce35fbcdec42"}

fake_responses = {
    '/v1/accelerators':
    {
        'GET': (
            {},
            {'accelerators': [ACCELERATOR1, ACCELERATOR2]},
        )
    }
}


class AcceleratorManagerTest(testtools.TestCase):

    def setUp(self):
        super(AcceleratorManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = accelerators.AcceleratorManager(self.api)

    def test_accelerators_list(self):
        accelerators = self.mgr.list()
        expect = [
            ('GET', '/v1/accelerators', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(accelerators, matchers.HasLength(2))



#   Copyright 2016 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import logging

from oslo_serialization import jsonutils

from cyborgclient import exceptions as exc

LOG = logging.getLogger(__name__)


def common_filters(marker=None, limit=None, sort_key=None, sort_dir=None):
    """Generate common filters for any list request.

    :param marker: entity ID from which to start returning entities.
    :param limit: maximum number of entities to return.
    :param sort_key: field to use for sorting.
    :param sort_dir: direction of sorting: 'asc' or 'desc'.
    :returns: list of string filters.
    """
    filters = []
    if isinstance(limit, int):
        filters.append('filters.field=limit')
        filters.append('filters.value=%d' % limit)
    if marker is not None:
        filters.append('filters.field=marker')
        filters.append('filters.value=%s' % marker)
    if sort_key is not None:
        filters.append('filters.field=sort_key')
        filters.append('filters.value=%s' % sort_key)
    if sort_dir is not None:
        filters.append('filters.field=sort_dir')
        filters.append('filters.value=%s' % sort_dir)
    return filters


def add_filters(filters, **kwargs):
    if kwargs:
        for field, value in kwargs.iteritems():
            filters.append('filters.field=%s' % field)
            filters.append('filters.value=%s' % value)
    return filters


def print_list_field(field):
    return lambda obj: ', '.join(getattr(obj, field))


def get_response_body(resp):
    body = resp.content
    content_type = resp.headers.get('Content-Type', '')
    if 'application/json' in content_type:
        try:
            body = resp.json()
        except ValueError:
            LOG.error('Could not decode response body as JSON')
    elif 'application/octet-stream' in content_type:
        try:
            body = resp.body()
        except ValueError:
            LOG.error('Could not decode response body as raw')
    else:
        body = None
    return body


def addresses_formatter(network_client, networks):
    output = []
    for (network, addresses) in networks.items():
        if not addresses:
            continue
        addrs = [addr['addr'] for addr in addresses]
        network_data = network_client.find_network(
            network, ignore_missing=False)
        net_ident = network_data.name or network_data.id
        addresses_csv = ', '.join(addrs)
        group = "%s=%s" % (net_ident, addresses_csv)
        output.append(group)
    return '; '.join(output)


def image_formatter(image_client, image_id):
    if image_id:
        image = image_client.images.get(image_id)
        return '%s (%s)' % (image.name, image_id)
    return ''


def flavor_formatter(bc_client, flavor_id):
    if flavor_id:
        flavor = bc_client.flavor.get(flavor_id)
        return '%s (%s)' % (flavor.name, flavor_id)
    return ''


def clean_listing_columns(headers, columns, data_sample):
    col_headers = []
    cols = []
    for header, col in zip(headers, columns):
        if hasattr(data_sample, col):
            col_headers.append(header)
            cols.append(col)
    return tuple(col_headers), tuple(cols)


def json_formatter(js):
    return jsonutils.dumps(js, indent=2, ensure_ascii=False)


def split_and_deserialize(string):
    """Split and try to JSON deserialize a string.

    Gets a string with the KEY=VALUE format, split it (using '=' as the
    separator) and try to JSON deserialize the VALUE.
    :returns: A tuple of (key, value).
    """

    try:
        key, value = string.split("=", 1)
    except ValueError:
        raise exc.CommandError(_('Attributes must be a list of '
                                 'PATH=VALUE not "%s"') % string)
    try:
        value = jsonutils.loads(value)
    except ValueError:
        pass

    return (key, value)


def args_array_to_patch(op, attributes):
    patch = []
    for attr in attributes:
        # Sanitize
        if not attr.startswith('/'):
            attr = '/' + attr

        if op in ['add', 'replace']:
            path, value = split_and_deserialize(attr)
            patch.append({'op': op, 'path': path, 'value': value})

        elif op == "remove":
            # For remove only the key is needed
            patch.append({'op': op, 'path': attr})
        else:
            raise exc.CommandError(_('Unknown PATCH operation: %s') % op)
    return patch



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import copy

from cyborgclient.osc.v2 import deployable as osc_deployable
from cyborgclient.tests.unit.osc.v2 import fakes as acc_fakes


class TestDeployable(acc_fakes.TestAccelerator):

    def setUp(self):
        super(TestDeployable, self).setUp()

        self.mock_acc_client = self.app.client_manager.accelerator
        self.mock_acc_client.reset_mock()


class TestDeployableList(TestDeployable):

    def setUp(self):
        super(TestDeployableList, self).setUp()

        self.mock_acc_client.deployables.return_value = [
            acc_fakes.FakeAcceleratorResource(
                None,
                copy.deepcopy(acc_fakes.DEPLOYABLE),
                loaded=True)
        ]

        self.cmd = osc_deployable.ListDeployable(self.app, None)

    def test_deployable_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        kwargs = {}

        self.mock_acc_client.deployables.assert_called_with(**kwargs)

        collist = (
            'uuid',
            'name',
            'device_id'
        )
        self.assertEqual(collist, columns)

        datalist = [(
            acc_fakes.deployable_uuid,
            acc_fakes.deployable_name,
            acc_fakes.deployable_device_id,
        ), ]
        self.assertEqual(datalist, list(data))

    def test_deployable_list_long(self):
        arglist = ['--long']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        kwargs = {}

        self.mock_acc_client.deployables.assert_called_with(**kwargs)

        collist = (
            'created_at',
            'updated_at',
            'uuid',
            'parent_id',
            'root_id',
            'name',
            'num_accelerators',
            'device_id'
        )
        self.assertEqual(collist, columns)

        datalist = [(
            acc_fakes.deployable_created_at,
            acc_fakes.deployable_updated_at,
            acc_fakes.deployable_uuid,
            acc_fakes.deployable_parent_id,
            acc_fakes.deployable_root_id,
            acc_fakes.deployable_name,
            acc_fakes.deployable_num_accelerators,
            acc_fakes.deployable_device_id,
        ), ]
        self.assertEqual(datalist, list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Cyborg v2 Acceleration accelerator action implementations"""
import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import utils as oscutils

from cyborgclient.common import utils
from cyborgclient import exceptions as exc
from cyborgclient.i18n import _


class ListDevice(command.Lister):
    """List all devices"""

    def get_parser(self, prog_name):
        parser = super(ListDevice, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("List additional fields in output")
        )
        return parser

    def take_action(self, parsed_args):
        acc_client = self.app.client_manager.accelerator

        if parsed_args.detail:
            column_headers = (
                "created_at",
                "updated_at",
                "uuid",
                "type",
                "vendor",
                "model",
                "hostname",
                "std_board_info",
                "vendor_board_info"
            )

            columns = (
                "created_at",
                "updated_at",
                "uuid",
                "type",
                "vendor",
                "model",
                "hostname",
                "std_board_info",
                "vendor_board_info"
            )
        else:
            column_headers = (
                "uuid",
                "type",
                "vendor",
                "hostname",
                "std_board_info",
            )
            columns = (
                "uuid",
                "type",
                "vendor",
                "hostname",
                "std_board_info",
            )

        data = acc_client.devices()
        if not data:
            return (), ()
        formatters = {}
        return (column_headers,
                (oscutils.get_item_properties(
                    s, columns, formatters=formatters) for s in data))


class ShowDevice(command.ShowOne):
    """Show device details."""
    log = logging.getLogger(__name__ + ".ShowDevice")

    def get_parser(self, prog_name):
        parser = super(ShowDevice, self).get_parser(prog_name)
        parser.add_argument(
            "device",
            metavar="<uuid>",
            help=_("UUID of the device.")
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        acc_client = self.app.client_manager.accelerator
        return _show_device(acc_client,
                            parsed_args.device)


def _show_device(acc_client, uuid):
    """Show detailed info about device."""
    columns = (
        "created_at",
        "updated_at",
        "uuid",
        "type",
        "vendor",
        "model",
        "hostname",
        "std_board_info",
        "vendor_board_info"
    )
    try:
        device = acc_client.get_device(uuid)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('device not found: %s') % uuid)
    formatters = {
        'data': utils.json_formatter,
    }
    data = device.to_dict()
    return columns, oscutils.get_dict_properties(data, columns,
                                                 formatters=formatters)



# Copyright (c) 2015 Thales Services SAS
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from cyborgclient.v1 import client


class ClientTest(testtools.TestCase):

    @mock.patch('cyborgclient.common.httpclient.SessionClient')
    @mock.patch('keystoneauth1.session.Session')
    def test_init_with_session(self, mock_session, http_client):
        session = mock.Mock()
        client.Client(session=session)
        mock_session.assert_not_called()
        http_client.assert_called_once_with(
            interface='public',
            region_name=None,
            service_name=None,
            service_type='accelerator',
            session=session,
            endpoint_override=None,
            api_version=None)

    @mock.patch('cyborgclient.common.httpclient.SessionClient')
    @mock.patch('keystoneauth1.session.Session')
    def test_init_with_endpoint_override(self, mock_session, http_client):
        session = mock.Mock()
        client.Client(session=session, endpoint_override='cyborgurl')
        mock_session.assert_not_called()
        http_client.assert_called_once_with(
            interface='public',
            region_name=None,
            service_name=None,
            service_type='accelerator',
            session=session,
            endpoint_override='cyborgurl',
            api_version=None)

    @mock.patch('cyborgclient.common.httpclient.SessionClient')
    @mock.patch('keystoneauth1.session.Session')
    def test_init_with_cyborg_url_and_endpoint_override(self, mock_session,
                                                        http_client):
        session = mock.Mock()
        client.Client(session=session, cyborg_url='cyborgurl',
                      endpoint_override='cyborgurl')
        mock_session.assert_not_called()
        http_client.assert_called_once_with(
            interface='public',
            region_name=None,
            service_name=None,
            service_type='accelerator',
            session=session,
            endpoint_override='cyborgurl',
            api_version=None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import copy

from cyborgclient.osc.v2 import accelerator_request as osc_accelerator_request
from cyborgclient.tests.unit.osc.v2 import fakes as acc_fakes


class TestAcceleratorRequest(acc_fakes.TestAccelerator):

    def setUp(self):
        super(TestAcceleratorRequest, self).setUp()

        self.mock_acc_client = self.app.client_manager.accelerator
        self.mock_acc_client.reset_mock()


class TestAcceleratorRequestList(TestAcceleratorRequest):

    def setUp(self):
        super(TestAcceleratorRequestList, self).setUp()

        self.mock_acc_client.accelerator_requests.return_value = [
            acc_fakes.FakeAcceleratorResource(
                None,
                copy.deepcopy(acc_fakes.ACCELERATOR_REQUEST),
                loaded=True)
        ]

        self.cmd = osc_accelerator_request.ListAcceleratorRequest(
            self.app, None
        )

    def test_accelerator_request_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        kwargs = {}

        self.mock_acc_client.accelerator_requests.assert_called_with(**kwargs)

        collist = (
            'uuid',
            'state',
            'device_profile_name',
            'instance_uuid',
            'attach_handle_type',
            'attach_handle_info',
        )
        self.assertEqual(collist, columns)

        datalist = [(
            acc_fakes.accelerator_request_uuid,
            acc_fakes.accelerator_request_state,
            acc_fakes.accelerator_request_device_profile_name,
            acc_fakes.accelerator_request_instance_uuid,
            acc_fakes.accelerator_request_attach_handle_type,
            acc_fakes.accelerator_request_attach_handle_info,
        ), ]
        self.assertEqual(datalist, list(data))

    def test_accelerator_request_list_long(self):
        arglist = ['--long']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        kwargs = {}

        self.mock_acc_client.accelerator_requests.assert_called_with(**kwargs)

        collist = (
            'uuid',
            'state',
            'device_profile_name',
            'hostname',
            'device_rp_uuid',
            'instance_uuid',
            'attach_handle_type',
            'attach_handle_info',
        )
        self.assertEqual(collist, columns)

        datalist = [(
            acc_fakes.accelerator_request_uuid,
            acc_fakes.accelerator_request_state,
            acc_fakes.accelerator_request_device_profile_name,
            acc_fakes.accelerator_request_hostname,
            acc_fakes.accelerator_request_device_rp_uuid,
            acc_fakes.accelerator_request_instance_uuid,
            acc_fakes.accelerator_request_attach_handle_type,
            acc_fakes.accelerator_request_attach_handle_info,
        ), ]
        self.assertEqual(datalist, list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Cyborg v2 Acceleration accelerator action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import utils as oscutils

from cyborgclient.common import utils
from cyborgclient import exceptions as exc
from cyborgclient.i18n import _


class ListAcceleratorRequest(command.Lister):
    """List all accelerator requests"""

    def get_parser(self, prog_name):
        parser = super(ListAcceleratorRequest, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("List additional fields in output")
        )
        return parser

    def take_action(self, parsed_args):
        acc_client = self.app.client_manager.accelerator

        if parsed_args.detail:
            column_headers = (
                "uuid",
                "state",
                "device_profile_name",
                "hostname",
                "device_rp_uuid",
                "instance_uuid",
                "attach_handle_type",
                "attach_handle_info",
            )
            columns = (
                "uuid",
                "state",
                "device_profile_name",
                "hostname",
                "device_rp_uuid",
                "instance_uuid",
                "attach_handle_type",
                "attach_handle_info",
            )
        else:
            column_headers = (
                "uuid",
                "state",
                "device_profile_name",
                "instance_uuid",
                "attach_handle_type",
                "attach_handle_info",
            )
            columns = (
                "uuid",
                "state",
                "device_profile_name",
                "instance_uuid",
                "attach_handle_type",
                "attach_handle_info",
            )

        data = acc_client.accelerator_requests()
        if not data:
            return (), ()
        formatters = {}
        return (column_headers,
                (oscutils.get_item_properties(
                    s, columns, formatters=formatters) for s in data))


class CreateAcceleratorRequest(command.ShowOne):
    """Register a new accelerator_request with the accelerator service"""

    log = logging.getLogger(__name__ + ".CreateAcceleratorRequest")

    def get_parser(self, prog_name):
        parser = super(CreateAcceleratorRequest, self).get_parser(prog_name)

        parser.add_argument(
            'device_profile_name',
            metavar='<device_profile_name>',
            help=_("The name of device_profile for accelerator_request."))
        parser.add_argument(
            'device_profile_group_id',
            metavar='<device_profile_group_id>',
            help=_("The group id of device_profile \
                   for the accelerator_request."))
        parser.add_argument(
            '--image-uuid',
            metavar='<glance_image_uuid>',
            dest='img_uuid',
            help=_("The uuid of image saved in glance."))

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        attrs = {
            'device_profile_name': parsed_args.device_profile_name,
            'device_profile_group_id': parsed_args.device_profile_group_id,
            'image_uuid': parsed_args.img_uuid,
        }

        accelerator_request = acc_client.create_accelerator_request(**attrs)

        return _show_accelerator_request(acc_client,
                                         accelerator_request.uuid)


class DeleteAcceleratorRequest(command.Command):
    """Delete accelerator request(s)."""

    log = logging.getLogger(__name__ + ".DeleteAcceleratorRequest")

    def get_parser(self, prog_name):
        parser = super(DeleteAcceleratorRequest, self).get_parser(prog_name)
        parser.add_argument(
            "accelerator_requests",
            metavar="<uuid>",
            nargs="+",
            help=_("UUID(s) of the accelerator_request(s) to delete.")
        )

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        failures = []
        for uuid in parsed_args.accelerator_requests:
            try:
                acc_client.delete_accelerator_request(uuid, False)
                print(_('Deleted accelerator_request %s') % uuid)
            except exc.ClientException as e:
                failures.append(_("Failed to delete accelerator_request\
                                %(accelerator_request)s: %(error)s")
                                % {'uuid': uuid, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class ShowAcceleratorRequest(command.ShowOne):
    """Show accelerator_request details."""

    log = logging.getLogger(__name__ + ".ShowAcceleratorRequest")

    def get_parser(self, prog_name):
        parser = super(ShowAcceleratorRequest, self).get_parser(prog_name)
        parser.add_argument(
            "accelerator_request",
            metavar="<uuid>",
            help=_("UUID of the accelerator_request.")
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        return _show_accelerator_request(acc_client,
                                         parsed_args.accelerator_request)


def _show_accelerator_request(acc_client, uuid):
    """Show detailed info about accelerator_request."""

    columns = (
        "uuid",
        "state",
        "device_profile_name",
        "hostname",
        "device_rp_uuid",
        "instance_uuid",
        "attach_handle_type",
        "attach_handle_info",
    )

    try:
        accelerator_request = acc_client.get_accelerator_request(uuid)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('accelerator_request not found: %s') % uuid)

    formatters = {
        'data': utils.json_formatter,
    }
    data = accelerator_request.to_dict()
    return columns, oscutils.get_dict_properties(data, columns,
                                                 formatters=formatters)


class BindAcceleratorRequest(command.ShowOne):
    """Bind accelerator to instance."""

    log = logging.getLogger(__name__ + ".BindAcceleratorRequest")

    def get_parser(self, prog_name):
        parser = super(BindAcceleratorRequest, self).get_parser(prog_name)

        parser.add_argument(
            'accelerator_request',
            metavar='<accelerator_request>',
            help=_("UUID of the accelerator request")
        )
        parser.add_argument(
            'hostname',
            metavar='<hostname>',
            help=_("Bind hostname of the accelerator request")
        )
        parser.add_argument(
            "instance_uuid",
            metavar="<instance_uuid>",
            help=_("Bind instance_uuid of the accelerator request")
        )
        parser.add_argument(
            "device_rp_uuid",
            metavar="<device_rp_uuid>",
            help=_("Bind device_rp_uuid of the accelerator request")
        )

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        properties = []
        if parsed_args.hostname:
            hostname = ["hostname=%s" % parsed_args.hostname]
            properties.extend(utils.args_array_to_patch('add', hostname))
        if parsed_args.instance_uuid:
            instance_uuid = ["instance_uuid=%s" % parsed_args.instance_uuid]
            properties.extend(utils.args_array_to_patch('add', instance_uuid))
        if parsed_args.device_rp_uuid:
            device_rp_uuid = ["device_rp_uuid=%s" % parsed_args.device_rp_uuid]
            properties.extend(utils.args_array_to_patch('add', device_rp_uuid))

        if properties:
            acc_client.update_accelerator_request(
                parsed_args.accelerator_request, properties)
            return _show_accelerator_request(acc_client,
                                             parsed_args.accelerator_request)
        else:
            self.log.warning("Please specify what to set.")


class UnbindAcceleratorRequest(command.ShowOne):
    """Unbind accelerator from instance."""

    log = logging.getLogger(__name__ + ".UnbindAcceleratorRequest")

    def get_parser(self, prog_name):
        parser = super(UnbindAcceleratorRequest, self).get_parser(prog_name)

        parser.add_argument(
            'accelerator_request',
            metavar='<accelerator_request>',
            help=_("UUID of the accelerator request")
        )

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        acc_client = self.app.client_manager.accelerator

        properties = [{'path': '/hostname', 'op': 'remove'},
                      {'path': '/instance_uuid', 'op': 'remove'},
                      {'path': '/device_rp_uuid', 'op': 'remove'}]

        acc_client.update_accelerator_request(parsed_args.accelerator_request,
                                              properties)

        return _show_accelerator_request(acc_client,
                                         parsed_args.accelerator_request)
